import logging

import kubernetes
import kubernetes.client as client
from dagster import Array, Field, Noneable, StringSource, check
from dagster.builtins import String
from dagster.config.field_utils import Shape
from dagster.core.host_representation.grpc_server_registry import GrpcServerEndpoint
from dagster.serdes import ConfigurableClass
from dagster_k8s import K8sRunLauncher
from dagster_k8s.executor import K8sStepHandler
from dagster_k8s.job import DagsterK8sJobConfig
from kubernetes.client.rest import ApiException

from ..user_code_launcher import ReconcileUserCodeLauncher
from .utils import (
    SERVICE_PORT,
    construct_repo_location_deployment,
    construct_repo_location_service,
    get_resource_label_for_metadata,
    unique_resource_name,
    wait_for_deployment_complete,
)

DEPLOYMENT_TIMEOUT = 90  # Can take time to pull images
SERVER_TIMEOUT = 60


class K8sUserCodeLauncher(ReconcileUserCodeLauncher, ConfigurableClass):
    def __init__(
        self,
        dagster_home,
        instance_config_map,
        inst_data=None,
        namespace=None,
        kubeconfig_file=None,
        pull_policy=None,
        env_secrets=None,
        service_account_name=None,
        volume_mounts=None,
    ):
        self._inst_data = inst_data
        self._logger = logging.getLogger("K8sUserCodeLauncher")
        self._dagster_home = check.str_param(dagster_home, "dagster_home")
        self._instance_config_map = check.str_param(instance_config_map, "instance_config_map")
        self._namespace = namespace
        self._pull_policy = pull_policy
        self._env_secrets = check.opt_list_param(env_secrets, "env_secrets", of_type=str)
        self._service_account_name = check.opt_str_param(
            service_account_name, "service_account_name"
        )
        self._volume_mounts = check.opt_list_param(volume_mounts, "volume_mounts")

        if kubeconfig_file:
            kubernetes.config.load_kube_config(kubeconfig_file)
        else:
            kubernetes.config.load_incluster_config()

        self._api_instance = client.AppsV1Api(client.ApiClient())  # todo close this

        # NOTE: for better availability, we should scan for existing servers and use them
        # (may need to check versions)
        self._cleanup_servers()
        super(K8sUserCodeLauncher, self).__init__()

        self._launcher = K8sRunLauncher(
            dagster_home=self._dagster_home,
            instance_config_map=self._instance_config_map,
            postgres_password_secret=None,
            job_image=None,
            image_pull_policy=self._pull_policy,
            image_pull_secrets=None,
            service_account_name=self._service_account_name,
            env_config_maps=None,
            env_secrets=self._env_secrets,
            job_namespace=self._namespace,
        )

    def register_instance(self, instance):
        super().register_instance(instance)
        self._launcher.register_instance(instance)

    @property
    def inst_data(self):
        return self._inst_data

    @classmethod
    def config_type(cls):
        return {
            "dagster_home": Field(StringSource, is_required=True),
            "instance_config_map": Field(StringSource, is_required=True),
            "namespace": Field(StringSource, is_required=False, default_value="default"),
            "kubeconfig_file": Field(StringSource, is_required=False),
            "pull_policy": Field(StringSource, is_required=False, default_value="Always"),
            "env_secrets": Field(
                Noneable(Array(StringSource)),
                is_required=False,
                description="A list of custom Secret names from which to draw environment "
                "variables (using ``envFrom``) for the Job. Default: ``[]``. See:"
                "https://kubernetes.io/docs/tasks/inject-data-application/distribute-credentials-secure/#configure-all-key-value-pairs-in-a-secret-as-container-environment-variables",
            ),
            "service_account_name": Field(
                Noneable(StringSource),
                is_required=False,
                description="Override the name of the Kubernetes service account under "
                "which to run.",
            ),
            "volume_mounts": Field(
                Array(
                    Shape(
                        {
                            "name": String,
                            "path": String,
                            "sub_path": String,
                            "secret": String,
                        }
                    )
                ),
                is_required=False,
                default_value=[],
                description="Volume mounts to attach to k8s jobs",
            ),
        }

    @staticmethod
    def from_config_value(inst_data, config_value):
        return K8sUserCodeLauncher(
            dagster_home=config_value.get("dagster_home"),
            instance_config_map=config_value.get("instance_config_map"),
            inst_data=inst_data,
            namespace=config_value.get("namespace"),
            kubeconfig_file=config_value.get("kubeconfig_file"),
            pull_policy=config_value.get("pull_policy"),
            env_secrets=config_value.get("env_secrets", []),
            service_account_name=config_value.get("service_account_name"),
            volume_mounts=config_value.get("volume_mounts"),
        )

    def _create_deployment_endpoint(self, location_name, metadata):
        resource_name = unique_resource_name(location_name)

        try:
            api_response = self._api_instance.create_namespaced_deployment(
                self._namespace,
                construct_repo_location_deployment(
                    location_name,
                    resource_name,
                    metadata,
                    self._pull_policy,
                    self._env_secrets,
                    self._service_account_name,
                ),
            )
            self._logger.info("Created deployment: {}".format(api_response.metadata.name))
        except ApiException as e:
            self._logger.error(
                "Exception when calling AppsV1Api->create_namespaced_deployment: %s\n" % e
            )
            raise e

        try:
            api_response = client.CoreV1Api().create_namespaced_service(
                self._namespace,
                construct_repo_location_service(location_name, resource_name, metadata),
            )
            self._logger.info("Created service: {}".format(api_response.metadata.name))
        except ApiException as e:
            self._logger.error(
                "Exception when calling AppsV1Api->create_namespaced_service: %s\n" % e
            )
            raise e

        wait_for_deployment_complete(
            resource_name,
            self._namespace,
            self._logger,
            location_name,
            metadata,
            existing_pods=[],
            timeout=DEPLOYMENT_TIMEOUT,
        )
        server_id = self._wait_for_server(
            host=resource_name, port=SERVICE_PORT, timeout=SERVER_TIMEOUT
        )

        endpoint = GrpcServerEndpoint(
            server_id=server_id,
            host=resource_name,
            port=SERVICE_PORT,
            socket=None,
        )

        return endpoint

    def _get_existing_deployments(self, location_name, metadata):
        return self._api_instance.list_namespaced_deployment(
            self._namespace,
            label_selector=f"metadata_hash={get_resource_label_for_metadata(location_name, metadata)}",
        ).items

    def _add_server(self, location_name, metadata):
        # check if the container already exists, remove it if so (could happen
        # if a previous attempt to set up the server failed)
        existing_deployments = self._get_existing_deployments(location_name, metadata)
        for existing_deployment in existing_deployments:
            self._logger.info(
                "Removing existing deployment {resource_name} for location {location_name}".format(
                    resource_name=existing_deployment.metadata.name,
                    location_name=location_name,
                )
            )
            self._remove_deployment(existing_deployment)

        return self._create_deployment_endpoint(location_name, metadata)

    def _gen_update_server(self, location_name, old_metadata, new_metadata):
        existing_deployments = self._get_existing_deployments(location_name, old_metadata)
        updated_server = self._create_deployment_endpoint(location_name, new_metadata)
        yield updated_server

        for existing_deployment in existing_deployments:
            self._logger.info(
                "Removing old deployment {resource_name} for location {location_name}".format(
                    resource_name=existing_deployment.metadata.name,
                    location_name=location_name,
                )
            )
            self._remove_deployment(existing_deployment)

    def _remove_server(self, location_name, metadata):
        existing_deployments = self._get_existing_deployments(location_name, metadata)

        for existing_deployment in existing_deployments:
            self._remove_deployment(existing_deployment.metadata.name)

    def _remove_deployment(self, resource_name):
        self._api_instance.delete_namespaced_deployment(resource_name, self._namespace)
        client.CoreV1Api().delete_namespaced_service(resource_name, self._namespace)
        self._logger.info("Removed deployment and service: {}".format(resource_name))

    def _cleanup_servers(self):
        deployments = self._api_instance.list_namespaced_deployment(
            self._namespace,
            label_selector="managed_by=K8sUserCodeLauncher",
        ).items
        for deployment in deployments:
            self._api_instance.delete_namespaced_deployment(
                deployment.metadata.name, self._namespace
            )

        services = (
            client.CoreV1Api()
            .list_namespaced_service(
                self._namespace,
                label_selector="managed_by=K8sUserCodeLauncher",
            )
            .items
        )
        for service in services:
            client.CoreV1Api().delete_namespaced_service(service.metadata.name, self._namespace)

        self._logger.info(
            "Deleted deployments: {} and services: {}".format(
                ",".join([deployment.metadata.name for deployment in deployments]),
                ",".join([service.metadata.name for service in services]),
            )
        )

    def __exit__(self, exception_type, exception_value, traceback):
        self._launcher.dispose()
        self._cleanup_servers()

    def step_handler(self):
        return K8sStepHandler(
            job_config=DagsterK8sJobConfig(
                dagster_home=self._dagster_home,
                instance_config_map=self._instance_config_map,
                postgres_password_secret=None,
                job_image=None,
                image_pull_policy=self._pull_policy,
                image_pull_secrets=None,
                service_account_name=self._service_account_name,
                env_config_maps=None,
                env_secrets=self._env_secrets,
                volume_mounts=self._volume_mounts,
            ),
            job_namespace=self._namespace,
        )

    def run_launcher(self):
        return self._launcher



import click

from .daemon import run_command


def create_agent_cli_group():
    group = click.Group(name="agent")
    group.add_command(run_command)
    return group


agent_cli = create_agent_cli_group()



from contextlib import ExitStack, contextmanager

import click
from dagster.cli.workspace.dynamic_workspace import DynamicWorkspace
from dagster.daemon.controller import DagsterDaemonController
from dagster.utils.interrupts import capture_interrupts
from dagster_cloud.daemon.dagster_cloud_api_daemon import DagsterCloudApiDaemon
from dagster_cloud.instance import DagsterCloudAgentInstance


@click.command(
    name="run",
    help="Run the dagster-cloud agent",
)
def run_command():
    with capture_interrupts():
        with DagsterCloudAgentInstance.get() as instance:
            if instance.is_ephemeral:
                raise Exception(
                    "dagster-cloud-daemon can't run using an in-memory instance. Make sure "
                    "the DAGSTER_HOME environment variable has been set correctly and that "
                    "you have created a dagster.yaml file there."
                )

            with ExitStack() as stack:
                grpc_server_registry = stack.enter_context(instance.create_user_code_launcher())
                daemons = [stack.enter_context(DagsterCloudApiDaemon())]

                @contextmanager
                def gen_workspace(_instance):
                    with DynamicWorkspace(grpc_server_registry) as workspace:
                        yield workspace

                with DagsterDaemonController(instance, daemons, gen_workspace) as controller:
                    controller.check_daemon_loop()



from typing import Any, Dict

from dagster import default_executors, executor
from dagster.core.errors import DagsterInvariantViolationError

DAGSTER_CLOUD_EXECUTOR_NAME = "cloud_isolated_step"
DAGSTER_CLOUD_EXECUTOR_CONFIG_SCHEMA: Dict[str, Any] = {}


@executor(
    name=DAGSTER_CLOUD_EXECUTOR_NAME,
    config_schema=DAGSTER_CLOUD_EXECUTOR_CONFIG_SCHEMA,
)
def cloud_isolated_step_executor(_context):
    raise DagsterInvariantViolationError("This executor should never run within user code")


default_cloud_executors = default_executors + [cloud_isolated_step_executor]



from dagster import Field, check
from dagster.config.validate import process_config, resolve_to_config_type
from dagster.core.errors import DagsterInvalidConfigError, DagsterInvariantViolationError
from dagster.core.instance import DagsterInstance
from dagster.core.instance.config import config_field_for_configurable_class
from dagster.core.instance.ref import configurable_class_data

from ..storage.client import create_proxy_client, dagster_cloud_api_config


class DagsterCloudInstance(DagsterInstance):
    pass


class DagsterCloudAgentInstance(DagsterCloudInstance):
    def __init__(self, *args, dagster_cloud_api, user_code_launcher=None, **kwargs):
        super().__init__(*args, **kwargs)

        processed_api_config = process_config(
            resolve_to_config_type(dagster_cloud_api_config()),
            check.dict_param(dagster_cloud_api, "dagster_cloud_api"),
        )

        if not processed_api_config.success:
            raise DagsterInvalidConfigError(
                "Errors whilst loading dagster_cloud_api config",
                processed_api_config.errors,
                dagster_cloud_api,
            )

        self._dagster_cloud_api_config = processed_api_config.value

        self._user_code_launcher_data = (
            configurable_class_data(user_code_launcher) if user_code_launcher else None
        )

        self._graphql_client = None

    def create_graphql_client(self):
        return create_proxy_client(self.dagster_cloud_graphql_url, self._dagster_cloud_api_config)

    @property
    def graphql_client(self):
        if self._graphql_client == None:
            self._graphql_client = self.create_graphql_client()

        return self._graphql_client

    @property
    def dagster_cloud_url(self):
        return self._dagster_cloud_api_config["url"]

    @property
    def dagster_cloud_graphql_url(self):
        return f"{self.dagster_cloud_url}/graphql"

    @property
    def dagster_cloud_username(self):
        return self._dagster_cloud_api_config.get("username")

    @property
    def dagster_cloud_password(self):
        return self._dagster_cloud_api_config.get("password")

    def create_user_code_launcher(self):
        if not self._user_code_launcher_data:
            # This is a user facing error. We should have more actionable advice and link to docs here.
            raise DagsterInvariantViolationError(
                "User code launcher is not configured for DagsterCloudAgentInstance. "
                "Configure a user code launcher under the user_code_launcher: key in your dagster.yaml file."
            )

        user_code_launcher = self._user_code_launcher_data.rehydrate()
        user_code_launcher.register_instance(self)
        return user_code_launcher

    @staticmethod
    def get():  # pylint: disable=arguments-differ
        instance = DagsterInstance.get()
        if not isinstance(instance, DagsterCloudAgentInstance):
            raise DagsterInvariantViolationError(
                """
DagsterInstance.get() did not return a DagsterCloudAgentInstance. Make sure that your"
`dagster.yaml` file is correctly configured to include the following:
custom_instance_class:
  module: dagster_cloud.instance
  class: DagsterCloudAgentInstance
"""
            )
        return instance

    @classmethod
    def config_schema(cls):
        return {
            "dagster_cloud_api": Field(dagster_cloud_api_config(), is_required=True),
            "user_code_launcher": config_field_for_configurable_class(),
        }

    def get_required_daemon_types(self):
        from dagster_cloud.daemon.dagster_cloud_api_daemon import DagsterCloudApiDaemon

        return [DagsterCloudApiDaemon.daemon_type()]



import click
from dagster_cloud.daemon.cli import agent_cli
from dagster_cloud.workspace.cli import workspace_cli


def create_dagster_cloud_cli():
    commands = {
        "agent": agent_cli,
        "workspace": workspace_cli,
    }

    @click.group(commands=commands)
    def group():
        "CLI tools for working with dagster-cloud."

    return group


cli = create_dagster_cloud_cli()


def main():
    cli(obj={})  # pylint:disable=E1123



import json
from collections import defaultdict
from typing import Any, Callable, Dict, Iterable, List, Optional

from dagster import check
from dagster.core.definitions.events import AssetKey
from dagster.core.events import DagsterEvent, DagsterEventType
from dagster.core.events.log import EventLogEntry
from dagster.core.execution.stats import RunStepKeyStatsSnapshot, StepEventStatus
from dagster.core.storage.event_log.base import (
    EventLogRecord,
    EventLogStorage,
    EventsCursor,
    extract_asset_events_cursor,
)
from dagster.core.storage.pipeline_run import PipelineRunStatsSnapshot
from dagster.serdes import (
    ConfigurableClass,
    ConfigurableClassData,
    deserialize_json_to_dagster_namedtuple,
    serialize_dagster_namedtuple,
)
from dagster.utils import datetime_as_float
from dagster.utils.error import SerializableErrorInfo

from ..errors import GraphQLStorageError
from .queries import (
    DELETE_EVENTS_MUTATION,
    ENABLE_SECONDARY_INDEX_MUTATION,
    END_WATCH_MUTATION,
    GET_ALL_ASSET_KEYS_QUERY,
    GET_ALL_ASSET_TAGS_QUERY,
    GET_ASSET_EVENTS_QUERY,
    GET_ASSET_RUN_IDS_QUERY,
    GET_ASSET_TAGS_QUERY,
    GET_EVENT_RECORDS_QUERY,
    GET_LOGS_FOR_RUN_QUERY,
    GET_STATS_FOR_RUN_QUERY,
    GET_STEP_STATS_FOR_RUN_QUERY,
    HAS_ASSET_KEY_QUERY,
    IS_ASSET_AWARE_QUERY,
    IS_PERSISTENT_QUERY,
    REINDEX_MUTATION,
    STORE_EVENT_MUTATION,
    UPGRADE_EVENT_LOG_STORAGE_MUTATION,
    WATCH_MUTATION,
    WIPE_ASSET_MUTATION,
    WIPE_EVENT_LOG_STORAGE_MUTATION,
)


def _input_for_serializable_error_info(serializable_error_info: Optional[SerializableErrorInfo]):
    check.opt_inst_param(serializable_error_info, "serializable_error_info", SerializableErrorInfo)

    if serializable_error_info is None:
        return None

    return {
        "message": serializable_error_info.mesage,
        "stack": serializable_error_info.stack,
        "clsName": serializable_error_info.cls_name,
        "cause": _input_for_serializable_error_info(serializable_error_info.cause),
    }


def _input_for_dagster_event(dagster_event: Optional[DagsterEvent]):
    if dagster_event is None:
        return None

    return {
        "eventTypeValue": dagster_event.event_type_value,
        "pipelineName": dagster_event.pipeline_name,
        "stepHandleKey": dagster_event.step_handle.to_key() if dagster_event.step_handle else None,
        "solidHandleId": (
            dagster_event.solid_handle.to_string() if dagster_event.solid_handle else None
        ),
        "stepKindValue": dagster_event.step_kind_value,
        "loggingTags": (
            json.dumps(dagster_event.logging_tags) if dagster_event.logging_tags else None
        ),
        "eventSpecificData": (
            serialize_dagster_namedtuple(dagster_event.event_specific_data)
            if dagster_event.event_specific_data
            else None
        ),
        "message": dagster_event.message,
        "pid": dagster_event.pid,
        "stepKey": dagster_event.step_key,
    }


def _event_log_entry_from_graphql(graphene_event_log_entry: Dict) -> EventLogEntry:
    check.dict_param(graphene_event_log_entry, "graphene_event_log_entry")

    return EventLogEntry(
        error_info=(
            deserialize_json_to_dagster_namedtuple(
                check.str_elem(graphene_event_log_entry, "errorInfo")
            )
            if graphene_event_log_entry.get("errorInfo") is not None
            else None
        ),
        message=check.str_elem(graphene_event_log_entry, "message"),
        level=check.int_elem(graphene_event_log_entry, "level"),
        user_message=check.str_elem(graphene_event_log_entry, "userMessage"),
        run_id=check.str_elem(graphene_event_log_entry, "runId"),
        timestamp=check.float_elem(graphene_event_log_entry, "timestamp"),
        step_key=check.opt_str_elem(graphene_event_log_entry, "stepKey"),
        pipeline_name=check.opt_str_elem(graphene_event_log_entry, "pipelineName"),
        dagster_event=(
            deserialize_json_to_dagster_namedtuple(
                check.str_elem(graphene_event_log_entry, "dagsterEvent")
            )
            if graphene_event_log_entry.get("dagsterEvent") is not None
            else None
        ),
    )


def _get_events_cursor_input(events_cursor) -> Optional[Dict[str, Any]]:
    check.opt_inst_param(events_cursor, "events_cursor", EventsCursor)
    if events_cursor is None:
        return None

    return {
        "id": events_cursor.id,
        "runUpdatedAfter": datetime_as_float(events_cursor.run_updated_after)
        if events_cursor.run_updated_after
        else None,
    }


def _event_record_from_graphql(graphene_event_record: Dict) -> EventLogRecord:
    check.dict_param(graphene_event_record, "graphene_event_record")

    return EventLogRecord(
        storage_id=check.int_elem(graphene_event_record, "storageId"),
        event_log_entry=_event_log_entry_from_graphql(graphene_event_record["eventLogEntry"]),
    )


class GraphQLEventLogStorage(EventLogStorage, ConfigurableClass):
    def __init__(self, inst_data=None, override_graphql_client=None):
        """Initialize this class directly only for test. Use the ConfigurableClass machinery to
        init from instance yaml."""
        self._inst_data = check.opt_inst_param(inst_data, "inst_data", ConfigurableClassData)
        self._override_graphql_client = override_graphql_client

    @property
    def inst_data(self):
        return self._inst_data

    @classmethod
    def config_type(cls):
        return {}

    @staticmethod
    def from_config_value(inst_data: ConfigurableClassData, config_value):
        return GraphQLEventLogStorage(inst_data=inst_data)

    @property
    def _graphql_client(self):
        return (
            self._override_graphql_client
            if self._override_graphql_client
            else self._instance.graphql_client
        )

    def _execute_query(self, query, variables=None):
        res = self._graphql_client.execute(query, variable_values=variables)
        if "errors" in res:
            raise GraphQLStorageError(res)
        return res

    def get_logs_for_run(
        self, run_id: str, cursor: Optional[int] = -1, of_type: Optional[DagsterEventType] = None
    ) -> Iterable[EventLogEntry]:
        check.opt_inst_param(of_type, "of_type", DagsterEventType)
        res = self._execute_query(
            GET_LOGS_FOR_RUN_QUERY,
            variables={
                "runId": check.str_param(run_id, "run_id"),
                "cursor": check.int_param(cursor, "cursor"),
                "ofType": of_type.value if of_type else None,
            },
        )
        return [
            _event_log_entry_from_graphql(event)
            for event in res["data"]["eventLogs"]["getLogsForRun"]
        ]

    def get_stats_for_run(self, run_id: str) -> PipelineRunStatsSnapshot:
        res = self._execute_query(
            GET_STATS_FOR_RUN_QUERY, variables={"runId": check.str_param(run_id, "run_id")}
        )
        stats = res["data"]["eventLogs"]["getStatsForRun"]
        return PipelineRunStatsSnapshot(
            run_id=check.str_elem(stats, "runId"),
            steps_succeeded=check.int_elem(stats, "stepsSucceeded"),
            steps_failed=check.int_elem(stats, "stepsFailed"),
            materializations=check.int_elem(stats, "materializations"),
            expectations=check.int_elem(stats, "expectations"),
            enqueued_time=check.opt_float_elem(stats, "enqueuedTime"),
            launch_time=check.opt_float_elem(stats, "launchTime"),
            start_time=check.opt_float_elem(stats, "startTime"),
            end_time=check.opt_float_elem(stats, "endTime"),
        )

    def get_step_stats_for_run(
        self, run_id: str, step_keys: Optional[List[str]] = None
    ) -> List[RunStepKeyStatsSnapshot]:
        res = self._execute_query(
            GET_STEP_STATS_FOR_RUN_QUERY,
            variables={
                "runId": check.str_param(run_id, "run_id"),
                "stepKeys": check.opt_list_param(step_keys, "step_keys", of_type=str),
            },
        )
        step_stats = res["data"]["eventLogs"]["getStepStatsForRun"]
        return [
            RunStepKeyStatsSnapshot(
                run_id=check.str_elem(stats, "runId"),
                step_key=check.str_elem(stats, "stepKey"),
                status=(
                    getattr(StepEventStatus, check.opt_str_elem(stats, "status"))
                    if stats.get("status") is not None
                    else None
                ),
                start_time=check.opt_float_elem(stats, "startTime"),
                end_time=check.opt_float_elem(stats, "endTime"),
                materializations=[
                    deserialize_json_to_dagster_namedtuple(materialization)
                    for materialization in check.opt_list_elem(
                        stats, "materializations", of_type=str
                    )
                ],
                expectation_results=[
                    deserialize_json_to_dagster_namedtuple(expectation_result)
                    for expectation_result in check.opt_list_elem(
                        stats, "expectationResults", of_type=str
                    )
                ],
                attempts=check.opt_int_elem(stats, "attempts"),
            )
            for stats in step_stats
        ]

    def store_event(self, event: EventLogEntry):
        check.inst_param(event, "event", EventLogEntry)

        self._execute_query(
            STORE_EVENT_MUTATION,
            variables={
                "eventRecord": {
                    "errorInfo": _input_for_serializable_error_info(event.error_info),
                    "message": event.message,
                    "level": event.level,
                    "userMessage": event.user_message,
                    "runId": event.run_id,
                    "timestamp": event.timestamp,
                    "stepKey": event.step_key,
                    "pipelineName": event.pipeline_name,
                    "dagsterEvent": _input_for_dagster_event(event.dagster_event),
                }
            },
        )

    def delete_events(self, run_id: str):
        self._execute_query(
            DELETE_EVENTS_MUTATION, variables={"runId": check.str_param(run_id, "run_id")}
        )
        return

    def upgrade(self):
        return self._execute_query(UPGRADE_EVENT_LOG_STORAGE_MUTATION)

    def reindex(self, print_fn: Callable = lambda _: None, force: bool = False):
        return self._execute_query(
            REINDEX_MUTATION, variables={"force": check.bool_param(force, "force")}
        )

    def wipe(self):
        return self._execute_query(WIPE_EVENT_LOG_STORAGE_MUTATION)

    def watch(self, run_id: str, start_cursor: int, callback: Callable):
        # If we wanted to implement this for real we would have to figure out how to pass the
        # callback as a serialized ConfigurableClass or equiv. May be better just to raise the
        # NotImplementedError in this class and remove the GQL endpoint completely
        return self._execute_query(
            WATCH_MUTATION,
            variables={
                "runId": check.str_param(run_id, "run_id"),
                "startCursor": check.int_param(start_cursor, "start_cursor"),
                "callback": r"{}",
            },
        )

    def end_watch(self, run_id: str, handler: Callable):
        return self._execute_query(
            END_WATCH_MUTATION,
            variables={"runId": check.str_param(run_id, "run_id"), "handler": r"{}"},
        )

    def enable_secondary_index(self, name: str, run_id: str = None):
        return self._execute_query(
            ENABLE_SECONDARY_INDEX_MUTATION,
            variables={
                "name": check.str_param(name, "name"),
                "runId": check.str_param(run_id, "run_id"),
            },
        )

    @property
    def is_persistent(self) -> bool:
        res = self._execute_query(IS_PERSISTENT_QUERY)
        return res["data"]["eventLogs"]["isPersistent"]

    @property
    def is_asset_aware(self) -> bool:
        res = self._execute_query(IS_ASSET_AWARE_QUERY)
        return res["data"]["eventLogs"]["isAssetAware"]

    def get_event_records(
        self,
        after_cursor=None,
        limit=None,
        ascending=False,
        of_type=None,
    ):
        check.opt_inst_param(after_cursor, "after_cursor", EventsCursor)
        check.opt_int_param(limit, "limit")
        check.opt_bool_param(ascending, "ascending")
        check.opt_inst_param(of_type, "of_type", DagsterEventType)

        res = self._execute_query(
            GET_EVENT_RECORDS_QUERY,
            variables={
                "afterCursor": _get_events_cursor_input(after_cursor),
                "limit": limit,
                "ascending": ascending,
                "ofType": of_type.value if of_type else None,
            },
        )
        return [
            _event_record_from_graphql(result)
            for result in res["data"]["eventLogs"]["getEventRecords"]
        ]

    def has_asset_key(self, asset_key: AssetKey):
        check.inst_param(asset_key, "asset_key", AssetKey)

        res = self._execute_query(
            HAS_ASSET_KEY_QUERY, variables={"assetKey": asset_key.to_string()}
        )
        return res["data"]["eventLogs"]["hasAssetKey"]

    def all_asset_keys(self) -> Iterable[AssetKey]:
        res = self._execute_query(GET_ALL_ASSET_KEYS_QUERY)
        return [
            AssetKey.from_db_string(asset_key_string)
            for asset_key_string in res["data"]["eventLogs"]["getAllAssetKeys"]
        ]

    def all_asset_tags(self) -> Dict[AssetKey, Dict[str, str]]:
        res = self._execute_query(GET_ALL_ASSET_TAGS_QUERY)
        asset_tags: Dict[AssetKey, Dict[str, str]] = defaultdict(dict)
        for asset_tag_entry in res["data"]["eventLogs"]["getAllAssetTags"]:
            asset_key_string = asset_tag_entry["assetKey"]
            tag_key = asset_tag_entry["tagKey"]
            tag_value = asset_tag_entry["tagValue"]
            asset_tags[AssetKey.from_db_string(asset_key_string)][tag_key] = tag_value

        return asset_tags

    def get_asset_tags(self, asset_key: AssetKey) -> Dict[str, str]:
        res = self._execute_query(
            GET_ASSET_TAGS_QUERY, variables={"assetKey": asset_key.to_string()}
        )
        return {entry["key"]: entry["value"] for entry in res["data"]["eventLogs"]["getAssetTags"]}

    def get_asset_events(
        self,
        asset_key: AssetKey,
        partitions: List[str] = None,
        before_cursor: int = None,
        after_cursor: int = None,
        limit: int = None,
        ascending: bool = False,
        include_cursor: bool = False,
        before_timestamp: float = None,
        cursor: int = None,
    ) -> Iterable[EventLogEntry]:
        check.inst_param(asset_key, "asset_key", AssetKey)
        before_cursor = check.opt_int_param(before_cursor, "before_cursor")
        after_cursor = check.opt_int_param(after_cursor, "after_cursor")
        before_timestamp = check.opt_float_param(before_timestamp, "before_timestamp")
        cursor = check.opt_int_param(cursor, "cursor")

        before_cursor, after_cursor = extract_asset_events_cursor(
            cursor, before_cursor, after_cursor, ascending
        )

        res = self._execute_query(
            GET_ASSET_EVENTS_QUERY,
            variables={
                "assetKey": asset_key.to_string(),
                "partitions": check.opt_list_param(partitions, "partitions", of_type=str),
                "beforeTimestamp": before_timestamp,
                "beforeCursor": before_cursor,
                "afterCursor": after_cursor,
                "limit": check.opt_int_param(limit, "limit"),
                "ascending": check.bool_param(ascending, "ascending"),
            },
        )
        asset_records = res["data"]["eventLogs"]["getAssetEvents"]
        if include_cursor:
            return [
                (
                    asset_record["id"],
                    _event_log_entry_from_graphql(asset_record["eventLogEntry"]),
                )
                for asset_record in asset_records
            ]
        else:
            return [
                _event_log_entry_from_graphql(asset_record["eventLogEntry"])
                for asset_record in asset_records
            ]

    def get_asset_run_ids(self, asset_key: AssetKey) -> Iterable[str]:
        res = self._execute_query(
            GET_ASSET_RUN_IDS_QUERY, variables={"assetKey": asset_key.to_string()}
        )
        return res["data"]["eventLogs"]["getAssetRunIds"]

    def wipe_asset(self, asset_key: AssetKey):
        res = self._execute_query(
            WIPE_ASSET_MUTATION, variables={"assetKey": asset_key.to_string()}
        )
        return res



ALL_STORED_JOB_STATE_QUERY = """
    query jobStates($repositoryOriginId: String, $jobType: JobType) {
        schedules {
            jobStates(repositoryOriginId: $repositoryOriginId, jobType: $jobType)
        }
    }
"""


ADD_JOB_STATE_MUTATION = """
    mutation addJobStateMutation($serializedJobState: String!) {
        schedules {
            addJobState(serializedJobState: $serializedJobState) {
                ok
            }
        }
    }
"""

CREATE_JOB_TICK_MUTATION = """
    mutation createJobTickMutation($serializedJobTickData: String!) {
        schedules {
            createJobTick(serializedJobTickData: $serializedJobTickData) {
                tickId
            }
        }
    }
"""

UPDATE_JOB_TICK_MUTATION = """
    mutation updateJobTickMutation($tickId: Int! $serializedJobTickData: String!) {
        schedules {
            updateJobTick(tickId: $tickId, serializedJobTickData: $serializedJobTickData) {
                ok
            }
        }
    }
"""

GET_JOB_TICK_STATS_QUERY = """
    query jobTickStats($jobOriginId: String!) {
        schedules {
            jobTickStats(jobOriginId: $jobOriginId) {
                ticksStarted
                ticksSucceeded
                ticksSkipped
                ticksFailed
            }
        }
    }

"""


GET_JOB_STATE_QUERY = """
    query jobState($jobOriginId: String!) {
        schedules {
            jobState(jobOriginId: $jobOriginId)
        }
    }
"""

GET_JOB_TICKS_QUERY = """
    query jobTicks($jobOriginId: String!, $before: Float, $after: Float, $limit: Int) {
        schedules {
            jobTicks(jobOriginId: $jobOriginId, before: $before, after: $after, limit: $limit)
        }
    }
"""

LATEST_JOB_TICK_QUERY = """
    query latestJobTick($jobOriginId: String!) {
        schedules {
            latestJobTick(jobOriginId: $jobOriginId)
        }
    }
"""

UPDATE_JOB_STATE_MUTATION = """
    mutation updateJobStateMutation($serializedJobState: String!) {
        schedules {
            updateJobState(serializedJobState: $serializedJobState) {
                ok
            }
        }
    }
"""

PURGE_JOB_TICKS_MUTATION = """
    mutation purgeJobTicksMutation($jobOriginId: String!, $tickStatus: JobTickStatus, $before: Float) {
        schedules {
            purgeJobTicks(jobOriginId: $jobOriginId, tickStatus: $tickStatus, before: $before) {
                ok
            }
        }
    }

"""



from typing import Any, Dict

from dagster import Field, IntSource, Noneable, Permissive, Selector, StringSource, check
from gql import Client, gql
from gql.transport.requests import RequestsHTTPTransport
from requests.auth import HTTPBasicAuth

from .errors import GraphQLStorageError


class GqlShimClient:
    """Adapter between the gql.Client and graphene.test.Client interfaces.

    Provides a uniform signature for execute(), uniform nesting of the return value, and a
    common error class."""

    def __init__(self, client: Client):
        self._client = check.inst_param(client, "client", Client)

    def execute(self, query: str, variable_values: Dict[str, Any] = None):
        try:
            return {"data": self._client.execute(gql(query), variable_values=variable_values)}
        except Exception as exc:  # pylint: disable=broad-except
            raise GraphQLStorageError(exc.__str__()) from exc


def create_proxy_client(url: str, config_value: Dict[str, Any], fetch_schema_from_transport=True):
    auth_config = config_value["auth"]

    # Basic auth only for now
    if "none" in auth_config:
        auth = HTTPBasicAuth("", "")
    else:
        auth = HTTPBasicAuth(auth_config["basic"]["username"], auth_config["basic"]["password"])

    transport = RequestsHTTPTransport(
        url=url,
        use_json=True,
        headers={"Content-type": "application/json", **config_value.get("headers", {})},
        verify=config_value.get("verify", True),
        retries=config_value.get("retries", 3),
        method=config_value.get("method", "POST"),
        timeout=config_value.get("timeout", None),
        cookies=config_value.get("cookies", {}),
        auth=auth,
    )
    return GqlShimClient(
        client=Client(transport=transport, fetch_schema_from_transport=fetch_schema_from_transport)
    )


def dagster_cloud_api_config():
    return {
        "url": Field(StringSource, default_value="http://localhost:2873"),
        "headers": Field(Permissive(), default_value={}),
        "cookies": Field(Permissive(), default_value={}),
        "auth": Field(
            Selector({"none": {}, "basic": {"username": StringSource, "password": StringSource}}),
            default_value={"none": {}},
        ),
        "timeout": Field(Noneable(IntSource), default_value=None),
        "verify": Field(bool, default_value=True),
        # This may be how we want to implement our retry policy, but it may be too restrictive:
        # we may want to put this logic into the storage itself so that we can do some kind of
        # logging
        "retries": Field(IntSource, default_value=3),
        "method": Field(StringSource, default_value="POST"),
    }



from typing import NamedTuple

from dagster import DagsterInstance, check
from dagster.core.executor.step_delegating import StepHandlerContext
from dagster.core.storage.pipeline_run import PipelineRun
from dagster.grpc.types import ExecuteStepArgs
from dagster.serdes.serdes import whitelist_for_serdes


@whitelist_for_serdes
class PersistedDagsterCloudStepHandlerContext(
    NamedTuple(
        "_PersistedDagsterCloudStepHandlerContext",
        [
            ("execute_step_args", ExecuteStepArgs),
            ("pipeline_run", PipelineRun),
        ],
    )
):
    def __new__(cls, execute_step_args: ExecuteStepArgs, pipeline_run: PipelineRun):
        return super(PersistedDagsterCloudStepHandlerContext, cls).__new__(
            cls,
            check.inst_param(execute_step_args, "execute_step_args", ExecuteStepArgs),
            check.inst_param(pipeline_run, "pipeline_run", PipelineRun),
        )


class DagsterCloudStepHandlerContext(StepHandlerContext):
    def serialize_for_cloud_api(self) -> PersistedDagsterCloudStepHandlerContext:
        # serialized contexts are passed to the user cloud, and should not contain host instances
        self._execute_step_args: ExecuteStepArgs = self._execute_step_args._replace(
            instance_ref=None
        )
        return PersistedDagsterCloudStepHandlerContext(
            execute_step_args=self.execute_step_args, pipeline_run=self.pipeline_run
        )

    @classmethod
    def deserialize(
        cls, instance: DagsterInstance, ctx_tuple: PersistedDagsterCloudStepHandlerContext
    ):
        execute_step_args = ctx_tuple.execute_step_args._replace(instance_ref=instance.get_ref())
        return cls(
            instance=instance,
            execute_step_args=execute_step_args,
            pipeline_run=ctx_tuple.pipeline_run,
        )

    @classmethod
    def from_oss_context(cls, context: StepHandlerContext):
        ctx = cls(
            instance=context.instance,
            execute_step_args=context.execute_step_args,
            pipeline_run=context.pipeline_run,
        )
        return ctx



ERROR_FRAGMENT = """
fragment errorFragment on PythonError {
  message
  className
  stack
  cause {
    message
    className
    stack
    cause {
      message
      className
      stack
    }
  }
}
"""

RUN_ROW_FRAGMENT = """
fragment runRecordFragment on RunRecord {
    storageId
    serializedPipelineRun
    createTimestamp
    updateTimestamp
}
"""

ADD_RUN_MUTATION = (
    ERROR_FRAGMENT
    + """
    mutation addRunMutation($serializedPipelineRun: String!) {
        runs {
            addRun(serializedPipelineRun: $serializedPipelineRun) {
                ok
                error {
                    ...errorFragment
                }
            }
        }
    }
"""
)

HANDLE_RUN_EVENT_MUTATION = """
    mutation handleRunEventMutation($runId: String!, $serializedEvent: String!) {
        runs {
            handleRunEvent(runId: $runId, serializedEvent: $serializedEvent) {
                ok
            }
        }
    }
"""

GET_RUNS_QUERY = """
    query getRunsQuery($filters: PipelineRunsFilter, $cursor: String, $limit: Int) {
        runs {
            getRuns(filters: $filters, cursor: $cursor, limit: $limit)
        }
    }
"""

GET_RUNS_COUNT_QUERY = """
    query getRunsCountQuery($filters: PipelineRunsFilter) {
        runs {
            getRunsCount(filters: $filters)
        }
    }
"""

GET_RUN_BY_ID_QUERY = """
    query getRunByIdQuery($runId: String!) {
        runs {
            getRunById(runId: $runId)
        }
    }
"""

HAS_RUN_QUERY = """
    query hasRunQuery($runId: String!) {
        runs {
            hasRun(runId: $runId)
        }
    }
"""

HAS_PIPELINE_SNAPSHOT_QUERY = """
    query hasPipelineSnapshotQuery($pipelineSnapshotId: String!) {
        runs {
            hasPipelineSnapshot(pipelineSnapshotId: $pipelineSnapshotId)
        }
    }
"""

GET_PIPELINE_SNAPSHOT_QUERY = """
    query getPipelineSnapshotQuery($pipelineSnapshotId: String!) {
        runs {
            getPipelineSnapshot(pipelineSnapshotId: $pipelineSnapshotId)
        }
    }
"""

GET_RUN_GROUP_QUERY = """
    query getRunGroupQuery($runId: String!) {
        runs {
            getRunGroup(runId: $runId) {
                rootRunId
                serializedRuns
            }
        }
    }
"""

GET_RUN_GROUPS_QUERY = """
    query getRunGroupsQuery($filters: PipelineRunsFilter, $cursor: String, $limit: Int) {
        runs {
            getRunGroups(filters: $filters, cursor: $cursor, limit: $limit) {
                rootRunId
                serializedRuns
                count
            }
        }
    }
"""

GET_RUN_RECORDS_QUERY = (
    RUN_ROW_FRAGMENT
    + """
    query getRunRecordsQuery($filters: PipelineRunsFilter, $limit: Int, $orderBy: String, $ascending: Boolean) {
        runs {
            getRunRecords(filters: $filters, limit: $limit, orderBy: $orderBy, ascending: $ascending) {
                ...runRecordFragment
            }
        }
    }
"""
)

GET_RUN_TAGS_QUERY = """
    query getRunTagsQuery {
        runs {
            getRunTags {
                key
                values
            }
        }
    }
"""

ADD_RUN_TAGS_MUTATION = """
    mutation addRunTagsMutation($runId: String!, $jsonNewTags: JSONString!) {
        runs {
            addRunTags(runId: $runId, jsonNewTags: $jsonNewTags) {
                ok
            }
        }
    }
"""

ADD_PIPELINE_SNAPSHOT_MUTATION = """
    mutation addPipelineSnapshotMutation($serializedPipelineSnapshot: String!) {
        runs {
            addPipelineSnapshot(serializedPipelineSnapshot: $serializedPipelineSnapshot) {
                ok
            }
        }
    }
"""

HAS_EXECUTION_PLAN_SNAPSHOT_QUERY = """
    query hasExecutionPlanSnapshotQuery($executionPlanSnapshotId: String!) {
        runs {
            hasExecutionPlanSnapshot(executionPlanSnapshotId: $executionPlanSnapshotId)
        }
    }
"""

ADD_EXECUTION_PLAN_SNAPSHOT_MUTATION = """
    mutation addExecutionPlanSnapshotMutation($serializedExecutionPlanSnapshot: String!) {
        runs {
            addExecutionPlanSnapshot(serializedExecutionPlanSnapshot: $serializedExecutionPlanSnapshot) {
                ok
            }
        }
    }
"""

GET_EXECUTION_PLAN_SNAPSHOT_QUERY = """
    query getExecutionPlanSnapshotQuery($executionPlanSnapshotId: String!) {
        runs {
            getExecutionPlanSnapshot(executionPlanSnapshotId: $executionPlanSnapshotId)
        }
    }
"""

ADD_DAEMON_HEARTBEAT_MUTATION = """
    mutation addDaemonHeartbeat($serializedDaemonHeartbeat: String!) {
        runs {
            addDaemonHeartbeat(serializedDaemonHeartbeat: $serializedDaemonHeartbeat) {
                ok
            }
        }
    }
"""

GET_DAEMON_HEARTBEATS_QUERY = """
    query getDaemonHeartbeatsQuery {
        runs {
            getDaemonHeartbeats
        }
    }
"""

GET_BACKFILLS_QUERY = """
    query getBackfillsQuery($status: String, $cursor: String, $limit: Int) {
        runs {
            getBackfills(status: $status, cursor: $cursor, limit: $limit)
        }
    }
"""

GET_BACKFILL_QUERY = """
    query getBackfillQuery($backfillId: String!) {
        runs {
            getBackfill(backfillId: $backfillId)
        }
    }
"""

ADD_BACKFILL_MUTATION = """
    mutation addBackfill($serializedPartitionBackfill: String!) {
        runs {
            addBackfill(serializedPartitionBackfill: $serializedPartitionBackfill) {
                ok
            }
        }
    }
"""

UPDATE_BACKFILL_MUTATION = """
    mutation updateBackfill($serializedPartitionBackfill: String!) {
        runs {
            updateBackfill(serializedPartitionBackfill: $serializedPartitionBackfill) {
                ok
            }
        }
    }
"""



from dagster import check
from dagster.config import Field, Shape
from dagster.config.source import StringSource
from dagster.config.validate import process_config


def process_workspace_config(workspace_config):
    check.dict_param(workspace_config, "workspace_config")
    check.invariant(isinstance(workspace_config.get("locations"), dict))

    for _, config in workspace_config["locations"].items():
        validation = process_config(LOCATION_CONFIG_SCHEMA, config)
        check.invariant(
            validation.success,
            ", ".join([error.message for error in validation.errors]),
        )

        python_file = config.get("python_file")
        package_name = config.get("package_name")
        check.invariant(
            (python_file or package_name) and not (python_file and package_name),
            "Must supply exactly one of a file name or a package name",
        )

    return True


LOCATION_CONFIG_SCHEMA = Shape(
    fields={
        "image": Field(config=StringSource, is_required=False),
        "python_file": Field(config=StringSource, is_required=False),
        "package_name": Field(config=StringSource, is_required=False),
    }
)
